""" Store package data in a SQL database """
from datetime import datetime

import logging
import transaction
from pkg_resources import parse_version
from sqlalchemy import engine_from_config, distinct, Column, DateTime, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
# pylint: disable=F0401,E0611
from zope.sqlalchemy import ZopeTransactionExtension
# pylint: enable=F0401,E0611

from .base import ICache
from pypicloud.models import Package
from sqlalchemy.types import TypeDecorator, TEXT
import json
from sqlalchemy.ext.mutable import Mutable


LOG = logging.getLogger(__name__)

Base = declarative_base()  # pylint: disable=C0103


class JSONEncodedDict(TypeDecorator):  # pylint: disable=W0223

    "Represents an immutable structure as a json-encoded string."

    impl = TEXT

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class MutableDict(Mutable, dict):

    """ SQLAlchemy dict field that tracks changes """

    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutableDict."

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."

        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."

        dict.__delitem__(self, key)
        self.changed()

MutableDict.associate_with(JSONEncodedDict)


class SQLPackage(Package, Base):

    """ Python package stored in SQLAlchemy """
    __tablename__ = 'packages'
    filename = Column(Text(), primary_key=True)
    name = Column(Text(), index=True, nullable=False)
    version = Column(Text(), nullable=False)
    last_modified = Column(DateTime(), index=True, nullable=False)
    data = Column(JSONEncodedDict(), nullable=False)


def create_schema(engine):
    """
    Create the database schema if needed

    Parameters
    ----------
    engine : :class:`sqlalchemy.Engine`

    Notes
    -----
    The method should only be called after importing all modules containing
    models which extend the ``Base`` object.

    """
    Base.metadata.create_all(bind=engine)


def drop_schema(engine):
    """
    Drop the database schema

    Parameters
    ----------
    engine : :class:`sqlalchemy.Engine`

    Notes
    -----
    The method should only be called after importing all modules containing
    models which extend the ``Base`` object.

    """
    Base.metadata.drop_all(bind=engine)


class SQLCache(ICache):

    """ Caching database that uses SQLAlchemy """
    package_class = SQLPackage

    def __init__(self, request=None, dbmaker=None, **kwargs):
        super(SQLCache, self).__init__(request, **kwargs)
        self.dbmaker = dbmaker
        self.db = self.dbmaker()

        if request is not None:
            def cleanup(_):
                """ Close the session after the request """
                self.db.close()
            request.add_finished_callback(cleanup)

    def reload_if_needed(self):
        super(SQLCache, self).reload_if_needed()
        if self.request is None:
            transaction.commit()
            self.db.close()

    @classmethod
    def configure(cls, settings):
        kwargs = super(SQLCache, cls).configure(settings)
        engine = engine_from_config(settings, prefix='db.')
        # Create SQL schema if not exists
        create_schema(engine)
        kwargs['dbmaker'] = sessionmaker(bind=engine,
                                         extension=ZopeTransactionExtension())
        return kwargs

    def fetch(self, filename):
        return self.db.query(SQLPackage).filter_by(filename=filename).first()

    def all(self, name):
        pkgs = self.db.query(SQLPackage).filter_by(name=name).all()
        pkgs.sort(reverse=True)
        return pkgs

    def distinct(self):
        names = self.db.query(distinct(SQLPackage.name))\
            .order_by(SQLPackage.name).all()
        return [n[0] for n in names]

    def summary(self):
        packages = {}
        for package in self.db.query(SQLPackage):
            pkg = packages.get(package.name)
            if pkg is None:
                pkg = {
                    'name': package.name,
                    'stable': None,
                    'unstable': '0',
                    'last_modified': datetime.fromtimestamp(0),
                }
                packages[package.name] = pkg
            if not package.is_prerelease:
                if pkg['stable'] is None:
                    pkg['stable'] = package.version
                else:
                    pkg['stable'] = max(pkg['stable'], package.version,
                                        key=parse_version)
            pkg['unstable'] = max(pkg['unstable'], package.version,
                                  key=parse_version)
            pkg['last_modified'] = max(pkg['last_modified'],
                                       package.last_modified)

        return packages.values()

    def clear(self, package):
        self.db.delete(package)

    def clear_all(self):
        # Release any transactions before we go reloading schema
        transaction.abort()
        engine = self.dbmaker.kw['bind']
        drop_schema(engine)
        create_schema(engine)

    def save(self, package):
        self.db.merge(package)



""" Base class for storage backends """
from pypicloud.models import Package


class IStorage(object):

    """ Base class for a backend that stores package files """

    def __init__(self, request):
        self.request = request

    @classmethod
    def configure(cls, settings):
        """ Configure the storage method with app settings """
        return {}

    def list(self, factory=Package):
        """ Return a list or generator of all packages """
        raise NotImplementedError

    def get_url(self, package):
        """
        Create or return an HTTP url for a package file

        By default this will return a link to the download endpoint

        /api/package/<package>/<filename>

        Returns
        -------
        link : str
            Link to the location of this package file

        """
        return self.request.app_url('api', 'package', package.name,
                                    package.filename)

    def download_response(self, package):
        """
        Return a HTTP Response that will download this package

        This is called from the download endpoint

        """
        raise NotImplementedError

    def upload(self, package, data):
        """
        Upload a package file to the storage backend

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`
            The package metadata
        data : file
            A file-like object that contains the package data

        """
        raise NotImplementedError

    def delete(self, package):
        """
        Delete a package file

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`
            The package metadata

        """
        raise NotImplementedError

    def open(self, package):
        """
        Get a buffer object that can read the package data

        This should be a context manager. It is used in migration scripts, not
        directly by the web application.

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        Examples
        --------
        ::

            with storage.open(package) as pkg_data:
                with open('outfile.tar.gz', 'w') as ofile:
                    ofile.write(pkg_data.read())

        """
        raise NotImplementedError



""" Tools and resources for traversal routing """
import functools


class IStaticResource(object):

    """ Simple resource base class for static-mapping of paths """
    __name__ = ''
    __parent__ = None
    subobjects = {}

    def __init__(self, request):
        self.request = request

    def __getitem__(self, name):
        child = self.subobjects[name](self.request)
        child.__parent__ = self
        child.__name__ = name
        return child


class IResourceFactory(object):

    """ Resource that generates child resources from a factory """
    __name__ = ''
    __parent__ = None
    __factory__ = lambda x: None

    def __init__(self, request):
        self.request = request

    def __getitem__(self, name):
        child = self.__factory__(name)
        child.__name__ = name
        child.__parent__ = self
        return child


class SimpleResource(object):

    """ Resource for simple pip calls """

    def __init__(self, request):
        self.request = request

    def __getitem__(self, name):
        child = SimplePackageResource(self.request, name)
        child.__parent__ = self
        child.__name__ = name
        return child


class SimplePackageResource(object):

    """ Resource for requesting simple endpoint package versions """

    __parent__ = None
    __name__ = None

    def __init__(self, request, name):
        self.request = request
        self.name = name
        self.__acl__ = request.access.get_acl(self.name)


class APIPackagingResource(IResourceFactory):

    """ Resource for api package queries """

    def __init__(self, request):
        super(APIPackagingResource, self).__init__(request)
        self.__factory__ = functools.partial(APIPackageResource, self.request)


class APIPackageResource(IResourceFactory):

    """ Resource for requesting package versions """

    def __init__(self, request, name):
        super(APIPackageResource, self).__init__(request)
        self.name = name
        self.__factory__ = functools.partial(
            APIPackageFileResource, self.request, self.name)
        self.__acl__ = request.access.get_acl(self.name)


class APIPackageFileResource(object):

    """ Resource for api endpoints dealing with a single package version """

    __parent__ = None
    __name__ = None

    def __init__(self, request, name, filename):
        self.request = request
        self.name = name
        self.filename = filename


class APIResource(IStaticResource):

    """ Resource for api calls """
    subobjects = {
        'package': APIPackagingResource,
    }


class AdminResource(IStaticResource):

    """ Resource for admin calls """


class Root(IStaticResource):

    """ Root context for PyPI Cloud """
    subobjects = {
        'api': APIResource,
        'admin': AdminResource,
        'simple': SimpleResource,
        'pypi': SimpleResource,
    }

    def __init__(self, request):
        super(Root, self).__init__(request)
        self.__acl__ = request.access.ROOT_ACL



""" Storage backend implementations """
from functools import partial

from .base import IStorage
from .files import FileStorage
from .s3 import S3Storage

from pyramid.path import DottedNameResolver


def get_storage_impl(settings):
    """ Get and configure the storage backend wrapper """
    resolver = DottedNameResolver(__name__)
    storage = settings.get('pypi.storage', 'file')
    if storage == 's3':
        storage = 'pypicloud.storage.S3Storage'
    elif storage == 'file':
        storage = 'pypicloud.storage.FileStorage'
    storage_impl = resolver.resolve(storage)
    kwargs = storage_impl.configure(settings)
    return partial(storage_impl, **kwargs)



""" Utilities """
import posixpath

import logging
from distlib.locators import Locator, SimpleScrapingLocator
from distlib.util import split_filename
from six.moves.urllib.parse import urlparse  # pylint: disable=F0401,E0611


LOG = logging.getLogger(__name__)
ALL_EXTENSIONS = Locator.source_extensions + Locator.binary_extensions


def parse_filename(filename, name=None):
    """ Parse a name and version out of a filename """
    version = None
    for ext in ALL_EXTENSIONS:
        if filename.endswith(ext):
            trimmed = filename[:-len(ext)]
            parsed_name, version = split_filename(trimmed, name)[:2]
            break
    if version is None:
        raise ValueError("Cannot parse package file '%s'" % filename)
    if name is None:
        name = parsed_name
    return normalize_name(name), version


def normalize_name(name):
    """ Normalize a python package name """
    return name.lower().replace('-', '_')


class BetterScrapingLocator(SimpleScrapingLocator):

    """ Layer on top of SimpleScrapingLocator that allows preferring wheels """
    prefer_wheel = True

    def __init__(self, *args, **kw):
        kw['scheme'] = 'legacy'
        super(BetterScrapingLocator, self).__init__(*args, **kw)

    def locate(self, requirement, prereleases=False, wheel=True):
        self.prefer_wheel = wheel
        return super(BetterScrapingLocator, self).locate(requirement, prereleases)

    def score_url(self, url):
        t = urlparse(url)
        filename = posixpath.basename(t.path)
        return (
            t.scheme == 'https',
            not (self.prefer_wheel ^ filename.endswith('.whl')),
            'pypi.python.org' in t.netloc,
            filename,
        )


def getdefaults(settings, *args):
    """
    Attempt multiple gets from a dict, returning a default value if none of the
    keys are found.

    """
    assert len(args) >= 3
    args, default = args[:-1], args[-1]
    canonical = args[0]
    for key in args:
        if key in settings:
            if key != canonical:
                LOG.warn("Using deprecated option '%s' "
                         "(replaced by '%s')", key, canonical)
            return settings[key]
    return default



""" Views for simple api calls that return json data """
import posixpath

import logging
import six
from contextlib import closing
from paste.httpheaders import CONTENT_DISPOSITION
from pyramid.httpexceptions import HTTPNotFound, HTTPForbidden, HTTPBadRequest
from pyramid.security import NO_PERMISSION_REQUIRED, remember
from pyramid.view import view_config
from pyramid_duh import argify, addslash
from six.moves.urllib.request import urlopen  # pylint: disable=F0401,E0611

from pypicloud.route import (APIResource, APIPackageResource,
                             APIPackagingResource, APIPackageFileResource)
from pypicloud.util import normalize_name


LOG = logging.getLogger(__name__)


@view_config(context=APIPackagingResource, request_method='GET',
             subpath=(), renderer='json')
@addslash
@argify
def all_packages(request, verbose=False):
    """ List all packages """
    if verbose:
        packages = request.db.summary()
    else:
        packages = request.db.distinct()
    i = 0
    while i < len(packages):
        package = packages[i]
        name = package if isinstance(package, basestring) else package['name']
        if not request.access.has_permission(name, 'read'):
            del packages[i]
            continue
        i += 1
    return {'packages': packages}


@view_config(context=APIPackageResource, request_method='GET',
             subpath=(), renderer='json', permission='read')
@addslash
def package_versions(context, request):
    """ List all unique package versions """
    normalized_name = normalize_name(context.name)
    versions = request.db.all(normalized_name)
    return {
        'packages': versions,
        'write': request.access.has_permission(normalized_name, 'write'),
    }


def fetch_dist(request, package_name, package_url):
    """ Fetch a Distribution and upload it to the storage backend """
    filename = posixpath.basename(package_url)
    with closing(urlopen(package_url)) as url:
        data = url.read()
    # TODO: digest validation
    return request.db.upload(filename, six.BytesIO(data), package_name), data


@view_config(context=APIPackageFileResource, request_method='GET',
             permission='read')
def download_package(context, request):
    """ Download package, or redirect to the download link """
    package = request.db.fetch(context.filename)
    if not package:
        if request.registry.fallback not in ['cache', 'mirror']:
            return HTTPNotFound()
        if not request.access.can_update_cache():
            return request.forbid()
        # If we are caching pypi, download the package from pypi and save it
        dists = request.locator.get_project(context.name)

        dist = None
        source_url = None
        for version, url_set in six.iteritems(dists.get('urls', {})):
            if dist is not None:
                break
            for url in url_set:
                if posixpath.basename(url) == context.filename:
                    source_url = url
                    dist = dists[version]
                    break
        if dist is None:
            return HTTPNotFound()
        LOG.info("Caching %s from %s", context.filename,
                 request.registry.fallback_url)
        package, data = fetch_dist(request, dist.name, source_url)
        disp = CONTENT_DISPOSITION.tuples(filename=package.filename)
        request.response.headers.update(disp)
        request.response.body = data
        request.response.content_type = 'application/octet-stream'
        return request.response
    response = request.db.download_response(package)
    return response


@view_config(context=APIPackageFileResource, request_method='POST',
             subpath=(), renderer='json', permission='write')
@argify
def upload_package(context, request, content):
    """ Upload a package """
    try:
        return request.db.upload(content.filename, content.file,
                                 name=context.name)
    except ValueError as e:  # pragma: no cover
        return HTTPBadRequest(*e.args)


@view_config(context=APIPackageFileResource, request_method='DELETE',
             subpath=(), permission='write')
def delete_package(context, request):
    """ Delete a package """
    package = request.db.fetch(context.filename)
    if package is None:
        return HTTPBadRequest("Could not find %s" % context.filename)
    request.db.delete(package)
    return request.response


@view_config(context=APIResource, name='user', request_method='PUT',
             subpath=('username/*'), renderer='json',
             permission=NO_PERMISSION_REQUIRED)
@argify
def register(request, password):
    """ Register a user """
    if not request.access.allow_register and not request.access.need_admin():
        return HTTPNotFound()
    username = request.named_subpaths['username']
    request.access.register(username, password)
    if request.access.need_admin():
        request.access.approve_user(username)
        request.access.set_user_admin(username, True)
        request.response.headers.extend(remember(request, username))
    return request.response


@view_config(context=APIResource, name='user', subpath=('password'),
             request_method='POST', permission='login')
@argify
def change_password(request, old_password, new_password):
    """ Change a user's password """
    if not request.access.verify_user(request.userid, old_password):
        return HTTPForbidden()
    request.access.edit_user_password(request.userid, new_password)
    return request.response


@view_config(context=APIResource, name='fetch', renderer='json',
             permission=NO_PERMISSION_REQUIRED)
@argify(wheel=bool, prerelease=bool)
def fetch_requirements(request, requirements, wheel=True, prerelease=False):
    """
    Fetch packages from the fallback_url

    Parameters
    ----------
    requirements : str
        Requirements in the requirements.txt format (with newlines)
    wheel : bool, optional
        If True, will prefer wheels (default True)
    prerelease : bool, optional
        If True, will allow prerelease versions (default False)

    Returns
    -------
    pkgs : list
        List of Package objects

    """
    if not request.access.can_update_cache():
        return HTTPForbidden()
    packages = []
    for line in requirements.splitlines():
        dist = request.locator.locate(line, prerelease, wheel)
        if dist is not None:
            try:
                packages.append(fetch_dist(request, dist.name, dist.source_url)[0])
            except ValueError:
                pass
    return {
        'pkgs': packages,
    }



""" Render views for logging in and out of the web interface """
from pyramid.httpexceptions import HTTPForbidden, HTTPFound, HTTPBadRequest
from pyramid.security import NO_PERMISSION_REQUIRED, remember, forget
from pyramid.view import view_config
from pyramid_duh import argify

from pypicloud.route import Root


@view_config(context=Root, name='login', request_method='GET',
             permission=NO_PERMISSION_REQUIRED, subpath=(),
             renderer='login.jinja2')
def get_login_page(request):
    """ Catch login and redirect to login wall """
    if request.userid is not None:
        # User is logged in and fetching /login, so redirect to /
        return HTTPFound(location=request.app_url())
    return {}


@view_config(context=HTTPForbidden, permission=NO_PERMISSION_REQUIRED,
             renderer='login.jinja2')
def do_forbidden(request):
    """ Intercept 403's and return 401's when necessary """
    return request.forbid()


@view_config(context=Root, name='login', request_method='POST', subpath=(),
             renderer='json', permission=NO_PERMISSION_REQUIRED)
@argify
def do_login(request, username, password):
    """ Check credentials and log in """
    if request.access.verify_user(username, password):
        request.response.headers.extend(remember(request, username))
        return {
            'next': request.app_url(),
        }
    else:
        return HTTPForbidden()


@view_config(context=Root, name='login', request_method='PUT', subpath=(),
             renderer='json', permission=NO_PERMISSION_REQUIRED)
@argify
def register(request, username, password):
    """ Check credentials and log in """
    if not request.access.allow_register():
        return HTTPForbidden()
    if request.access.user_data(username) is not None:
        return HTTPBadRequest()
    request.access.register(username, password)
    return request.response


@view_config(context=Root, name='logout', subpath=())
def logout(request):
    """ Delete the user session """
    request.response.headers.extend(forget(request))
    return HTTPFound(location=request.app_url())



""" API endpoints for admin controls """
import gzip
import json
from paste.httpheaders import CONTENT_DISPOSITION
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.response import FileIter
from pyramid.view import view_config, view_defaults
from pyramid_duh import argify
from six import BytesIO

from pypicloud.route import AdminResource


@view_defaults(context=AdminResource, subpath=(), permission='admin',
               renderer='json')
class AdminEndpoints(object):

    """ Collection of admin endpoints """

    def __init__(self, request):
        self.request = request

    @view_config(name='rebuild')
    def rebuild_package_list(self):
        """ Rebuild the package cache in the database """
        self.request.db.reload_from_storage()
        return self.request.response

    @view_config(name='pending_users', request_method='GET')
    def get_pending_users(self):
        """ Get the list of pending users """
        return self.request.access.pending_users()

    @view_config(name='user', request_method='GET')
    def get_users(self):
        """ Get the list of users """
        return self.request.access.user_data()

    @view_config(name='user', subpath=('username/*'), request_method='GET')
    def get_user(self):
        """ Get a single user """
        username = self.request.named_subpaths['username']
        return self.request.access.user_data(username)

    @view_config(name='user', subpath=('username/*'), request_method='DELETE')
    def delete_user(self):
        """ Delete a user """
        username = self.request.named_subpaths['username']
        self.request.access.delete_user(username)
        return self.request.response

    @view_config(name='user', subpath=('username/*', 'approve'),
                 request_method='POST')
    def approve_user(self):
        """ Approve a pending user """
        username = self.request.named_subpaths['username']
        self.request.access.approve_user(username)
        return self.request.response

    @view_config(name='user', subpath=('username/*', 'admin'),
                 request_method='POST')
    @argify
    def set_admin_status(self, admin):
        """ Approve a pending user """
        username = self.request.named_subpaths['username']
        self.request.access.set_user_admin(username, admin)
        return self.request.response

    @view_config(name='user', subpath=('username/*', 'group', 'group/*'),
                 request_method='PUT')
    @view_config(name='user', subpath=('username/*', 'group', 'group/*'),
                 request_method='DELETE')
    def mutate_group_member(self):
        """ Add a user to a group """
        username = self.request.named_subpaths['username']
        group = self.request.named_subpaths['group']
        self.request.access.edit_user_group(username, group,
                                            self.request.method == 'PUT')
        return self.request.response

    @view_config(name='group', request_method='GET')
    def get_groups(self):
        """ Get the list of groups """
        return self.request.access.groups()

    @view_config(name='group', subpath=('group/*'), request_method='PUT')
    def create_group(self):
        """ Create a group """
        group = self.request.named_subpaths['group']
        if group in ('everyone', 'authenticated'):
            return HTTPBadRequest("'%s' is a reserved name" % group)
        self.request.access.create_group(group)
        return self.request.response

    @view_config(name='group', subpath=('group/*'), request_method='DELETE')
    def delete_group(self):
        """ Delete a group """
        group = self.request.named_subpaths['group']
        self.request.access.delete_group(group)
        return self.request.response

    @view_config(name='user', subpath=('username/*', 'permissions'))
    def get_user_permissions(self):
        """ Get the package permissions for a user """
        username = self.request.named_subpaths['username']
        return self.request.access.user_package_permissions(username)

    @view_config(name='group', subpath=('group/*'))
    def get_group(self):
        """ Get the members and package permissions for a group """
        group = self.request.named_subpaths['group']
        return {
            'members': self.request.access.group_members(group),
            'packages': self.request.access.group_package_permissions(group),
        }

    @view_config(name='package', subpath=('package/*'), request_method='GET')
    def get_package_permissions(self):
        """ Get the user and group permissions set on a package """
        package = self.request.named_subpaths['package']
        user_perms = [{'username': key, 'permissions': val} for key, val in
                      self.request.access.user_permissions(package).iteritems()]
        group_perms = [{'group': key, 'permissions': val} for key, val in
                       self.request.access.group_permissions(package).iteritems()]
        return {
            'user': user_perms,
            'group': group_perms,
        }

    @view_config(name='package',
                 subpath=('package/*', 'type/user|group/r',
                          'name/*', 'permission/read|write/r'),
                 request_method='PUT')
    @view_config(name='package',
                 subpath=('package/*', 'type/user|group/r',
                          'name/*', 'permission/read|write/r'),
                 request_method='DELETE')
    def edit_permission(self):
        """ Edit user permission on a package """
        package = self.request.named_subpaths['package']
        name = self.request.named_subpaths['name']
        permission = self.request.named_subpaths['permission']
        owner_type = self.request.named_subpaths['type']
        if owner_type == 'user':
            self.request.access.edit_user_permission(
                package, name, permission, self.request.method == 'PUT')
        else:
            self.request.access.edit_group_permission(
                package, name, permission,
                self.request.method == 'PUT')
        return self.request.response

    @view_config(name='register', request_method='POST')
    @argify
    def toggle_allow_register(self, allow):
        """ Allow or disallow user registration """
        self.request.access.set_allow_register(allow)
        return self.request.response

    @view_config(name='acl.json.gz', request_method='GET')
    def download_access_control(self):
        """ Download the ACL data as a gzipped-json file """
        data = self.request.access.dump()
        compressed = BytesIO()
        zipfile = gzip.GzipFile(mode='wb', fileobj=compressed)
        json.dump(data, zipfile, separators=(',', ':'))
        zipfile.close()
        compressed.seek(0)

        disp = CONTENT_DISPOSITION.tuples(filename='acl.json.gz')
        self.request.response.headers.update(disp)
        self.request.response.app_iter = FileIter(compressed)
        return self.request.response



""" Backend that reads access control rules from config file """
import logging
from collections import defaultdict
from pyramid.security import Everyone, Authenticated
from pyramid.settings import aslist, asbool

from .base import IAccessBackend


LOG = logging.getLogger(__name__)


class ConfigAccessBackend(IAccessBackend):

    """ Access Backend that uses values set in the config file """

    def __init__(self, request=None, settings=None, admins=None,
                 group_map=None, user_groups=None, **kwargs):
        super(ConfigAccessBackend, self).__init__(request, **kwargs)
        self._settings = settings
        self.admins = admins
        self.group_map = group_map
        self.user_groups = user_groups

    @classmethod
    def configure(cls, settings):
        kwargs = super(ConfigAccessBackend, cls).configure(settings)
        if asbool(settings.get('auth.zero_security_mode', False)):
            LOG.warn("Using deprecated option 'auth.zero_security_mode' "
                     "(replaced by 'pypi.default_read' and "
                     "'pypi.default_write'")
            kwargs['default_read'] = [Everyone]
            kwargs['default_write'] = [Authenticated]
        kwargs['settings'] = settings
        kwargs['admins'] = aslist(settings.get('auth.admins', []))
        user_groups = defaultdict(list)
        group_map = {}

        # Build dict that maps users to list of groups
        for key, value in settings.iteritems():
            if not key.startswith('group.'):
                continue
            group_name = key[len('group.'):]
            members = aslist(value)
            group_map[group_name] = members
            for member in members:
                user_groups[member].append(group_name)
        kwargs['group_map'] = group_map
        kwargs['user_groups'] = user_groups
        return kwargs

    def _get_password_hash(self, username):
        key = "user.%s" % username
        return self._settings.get(key)

    def groups(self, username=None):
        if username is None:
            return self.group_map.keys()
        else:
            return self.user_groups[username]

    def group_members(self, group):
        return self.group_map.get(group, [])

    def is_admin(self, username):
        return username in self.admins

    @staticmethod
    def _perms_from_short(value):
        """ Convert a 'r' or 'rw' specification to a list of permissions """
        value = value or ''
        perms = []
        if 'r' in value:
            perms.append('read')
        if 'w' in value:
            perms.append('write')
        return perms

    def group_permissions(self, package, group=None):
        if group is not None:
            key = 'package.%s.group.%s' % (package, group)
            return self._perms_from_short(self._settings.get(key))
        perms = {}
        group_prefix = 'package.%s.group.' % package
        for key, value in self._settings.iteritems():
            if not key.startswith(group_prefix):
                continue
            group = key[len(group_prefix):]
            perms[group] = self._perms_from_short(value)
        return perms

    def user_permissions(self, package, username=None):
        if username is not None:
            key = 'package.%s.user.%s' % (package, username)
            return self._perms_from_short(self._settings.get(key))
        perms = {}
        user_prefix = 'package.%s.user.' % package
        for key, value in self._settings.iteritems():
            if not key.startswith(user_prefix):
                continue
            user = key[len(user_prefix):]
            perms[user] = self._perms_from_short(value)
        return perms

    def user_data(self, username=None):
        if username is not None:
            return {
                'username': username,
                'admin': self.is_admin(username),
                'groups': self.groups(username),
            }
        users = []
        user_prefix = 'user.'
        for key in self._settings:
            if not key.startswith(user_prefix):
                continue
            username = key[len(user_prefix):]
            users.append({
                'username': username,
                'admin': self.is_admin(username)
            })
        return users

    def user_package_permissions(self, username):
        perms = []
        for key, value in self._settings.iteritems():
            pieces = key.split('.')
            if (len(pieces) != 4 or pieces[0] != 'package' or
                    pieces[2] != 'user'):
                continue
            package, user = pieces[1], pieces[3]
            if user == username:
                perms.append({
                    'package': package,
                    'permissions': self._perms_from_short(value),

                })
        return perms

    def group_package_permissions(self, group):
        perms = []
        for key, value in self._settings.iteritems():
            pieces = key.split('.')
            if (len(pieces) != 4 or pieces[0] != 'package' or
                    pieces[2] != 'group'):
                continue
            package, groupname = pieces[1], pieces[3]
            if group == groupname:
                perms.append({
                    'package': package,
                    'permissions': self._perms_from_short(value),

                })
        return perms

    def load(self, data):
        lines = []
        admins = []
        for user in data['users']:
            lines.append('user.{username} = {password}'.format(**user))
            if user.get('admin'):
                admins.append(user['username'])

        if admins:
            lines.append('auth.admins =')
            for admin in admins:
                lines.append('    {0}'.format(admin))

        for group, members in data['groups'].iteritems():
            lines.append('group.{0} ='.format(group))
            for member in members:
                lines.append('    {0}'.format(member))

        def encode_permissions(perms):
            """ Encode a permissions list as the r/rw specification """
            ret = ''
            if 'read' in perms:
                ret += 'r'
            if 'write' in perms:
                ret += 'w'
            return ret

        for package, groups in data['packages']['groups'].iteritems():
            for group, permissions in groups.iteritems():
                lines.append('package.{0}.group.{1} = {2}'
                             .format(package, group,
                                     encode_permissions(permissions)))

        for package, users in data['packages']['users'].iteritems():
            for user, permissions in users.iteritems():
                lines.append('package.{0}.user.{1} = {2}'
                             .format(package, user,
                                     encode_permissions(permissions)))

        return '\n'.join(lines)



""" Model objects """
import re
from datetime import datetime

import pkg_resources

from .compat import total_ordering
from .util import normalize_name


@total_ordering
class Package(object):

    """
    Representation of a versioned package

    Parameters
    ----------
    name : str
        The name of the package (will be normalized)
    version : str
        The version number of the package
    filename : str
        The name of the package file
    last_modified : datetime, optional
        The datetime when this package was uploaded (default now)
    **kwargs : dict
        Metadata about the package

    """

    def __init__(self, name, version, filename, last_modified=None, **kwargs):
        self.name = normalize_name(name)
        self.version = version
        self._parsed_version = None
        self.filename = filename
        if last_modified is not None:
            self.last_modified = last_modified
        else:
            self.last_modified = datetime.utcnow()
        self.data = kwargs

    def get_url(self, request):
        """ Create path to the download link """
        return request.db.get_url(self)

    @property
    def parsed_version(self):
        """ Parse and cache the version using pkg_resources """
        # Use getattr because __init__ isn't called by some ORMs.
        if getattr(self, '_parsed_version', None) is None:
            self._parsed_version = pkg_resources.parse_version(self.version)
        return self._parsed_version

    @property
    def is_prerelease(self):
        """ Returns True if the version is a prerelease version """
        return re.match(r'^\d+(\.\d+)*$', self.version) is None

    def __hash__(self):
        return hash(self.name) + hash(self.version)

    def __eq__(self, other):
        return self.name == other.name and self.version == other.version

    def __lt__(self, other):
        return ((self.name, self.parsed_version) <
                (other.name, other.parsed_version))

    def __repr__(self):
        return unicode(self)

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return u'Package(%s)' % (self.filename)

    def __json__(self, request):
        return {
            'name': self.name,
            'filename': self.filename,
            'last_modified': self.last_modified,
            'version': self.version,
            'url': self.get_url(request),
        }



""" The access backend object base class """
from collections import defaultdict
from passlib.apps import custom_app_context as pwd_context
from pyramid.security import (Authenticated, Everyone,
                              effective_principals, Allow, Deny,
                              ALL_PERMISSIONS)
from pyramid.settings import aslist


def group_to_principal(group):
    """ Convert a group to its corresponding principal """
    if group in (Everyone, Authenticated) or group.startswith('group:'):
        return group
    elif group == 'everyone':
        return Everyone
    elif group == 'authenticated':
        return Authenticated
    else:
        return 'group:' + group


def groups_to_principals(groups):
    """ Convert a list of groups to a list of principals """
    return [group_to_principal(g) for g in groups]


class IAccessBackend(object):

    """ Base class for retrieving user and package permission data """

    mutable = False
    ROOT_ACL = [
        (Allow, Authenticated, 'login'),
        (Allow, 'admin', ALL_PERMISSIONS),
        (Deny, Everyone, ALL_PERMISSIONS),
    ]

    def __init__(self, request=None, default_read=None, default_write=None,
                 cache_update=None):
        self.request = request
        self.default_read = default_read
        self.default_write = default_write
        self.cache_update = cache_update

    @classmethod
    def configure(cls, settings):
        """ Configure the access backend with app settings """
        return {
            'default_read': aslist(settings.get('pypi.default_read',
                                                ['authenticated'])),
            'default_write': aslist(settings.get('pypi.default_write', [])),
            'cache_update': aslist(settings.get('pypi.cache_update',
                                                ['authenticated'])),
        }

    def allowed_permissions(self, package):
        """
        Get all allowed permissions for all principals on a package

        Returns
        -------
        perms : dict
            Mapping of principal to tuple of permissions

        """
        all_perms = {}
        for user, perms in self.user_permissions(package).iteritems():
            all_perms['user:' + user] = tuple(perms)

        for group, perms in self.group_permissions(package).iteritems():
            all_perms[group_to_principal(group)] = tuple(perms)

        # If there are no group or user specifications for the package, use the
        # default
        if len(all_perms) == 0:
            for principal in groups_to_principals(self.default_read):
                all_perms[principal] = ('read',)
            for principal in groups_to_principals(self.default_write):
                if principal in all_perms:
                    all_perms[principal] += ('write',)
                else:
                    all_perms[principal] = ('write',)
        return all_perms

    def get_acl(self, package):
        """ Construct an ACL for a package """
        acl = []
        permissions = self.allowed_permissions(package)
        for principal, perms in permissions.iteritems():
            for perm in perms:
                acl.append((Allow, principal, perm))
        return acl

    def has_permission(self, package, perm):
        """ Check if this user has a permission for a package """
        current_userid = self.request.userid
        if current_userid is not None and self.is_admin(current_userid):
            return True

        perms = self.allowed_permissions(package)
        for principal in effective_principals(self.request):
            if perm in perms.get(principal, []):
                return True
        return False

    def user_principals(self, username):
        """
        Get a list of principals for a user

        Parameters
        ----------
        username : str

        Returns
        -------
        principals : list

        """
        principals = ['user:' + username, Everyone, Authenticated]
        if self.is_admin(username):
            principals.append('admin')
        for group in self.groups(username):
            principals.append('group:' + group)
        return principals

    def in_group(self, username, group):
        """
        Find out if a user is in a group

        Parameters
        ----------
        username : str
            Name of user. May be None for the anonymous user.
        group : str
            Name of the group. Supports 'everyone', 'authenticated', and
            'admin'.

        Returns
        -------
        member : bool

        """
        if group in ('everyone', Everyone):
            return True
        elif username is None:
            return False
        elif group in ('authenticated', Authenticated):
            return True
        elif group == 'admin' and self.is_admin(username):
            return True
        else:
            return group in self.groups(username)

    def in_any_group(self, username, groups):
        """
        Find out if a user is in any of a set of groups

        Parameters
        ----------
        username : str
            Name of user. May be None for the anonymous user.
        groups : list
            list of group names. Supports 'everyone', 'authenticated', and
            'admin'.

        Returns
        -------
        member : bool

        """
        return any((self.in_group(username, group) for group in groups))

    def can_update_cache(self):
        """
        Return True if the user has permissions to update the pypi cache
        """
        return self.in_any_group(self.request.userid, self.cache_update)

    def need_admin(self):
        """
        Find out if there are any admin users

        This should only be overridden by mutable backends

        Returns
        -------
        need_admin : bool
            True if no admin user exists and the backend is mutable, False
            otherwise

        """
        return False

    def allow_register(self):
        """
        Check if the backend allows registration

        This should only be overridden by mutable backends

        Returns
        -------
        allow : bool

        """
        return False

    def verify_user(self, username, password):
        """
        Check the login credentials of a user

        For Mutable backends, pending users should fail to verify

        Parameters
        ----------
        username : str
        password : str

        Returns
        -------
        valid : bool
            True if user credentials are valid, false otherwise

        """
        stored_pw = self._get_password_hash(username)
        if self.mutable:
            # if a user is pending, user_data will be None
            user_data = self.user_data(username)
            if user_data is None:
                return False
        return bool(stored_pw and pwd_context.verify(password, stored_pw))

    def _get_password_hash(self, username):
        """ Get the stored password hash for a user """
        raise NotImplementedError

    def groups(self, username=None):
        """
        Get a list of all groups

        If a username is specified, get all groups that the user belongs to

        Parameters
        ----------
        username : str, optional

        Returns
        -------
        groups : list
            List of group names

        """
        raise NotImplementedError

    def group_members(self, group):
        """
        Get a list of users that belong to a group

        Parameters
        ----------
        group : str

        Returns
        -------
        users : list
            List of user names

        """
        raise NotImplementedError

    def is_admin(self, username):
        """
        Check if the user is an admin

        Parameters
        ----------
        username : str

        Returns
        -------
        is_admin : bool

        """
        raise NotImplementedError

    def group_permissions(self, package, group=None):
        """
        Get a mapping of all groups to their permissions on a package

        If a group is specified, just return the list of permissions for that
        group

        Parameters
        ----------
        package : str
            The name of a python package
        group : str, optional
            The name of a single group the check


        Returns
        -------
        permissions : dict
            If group is None, mapping of group name to a list of permissions
            (which can contain 'read' and/or 'write')
        permissions : list
            If group is not None, a list of permissions for that group

        Notes
        -----
        You may specify special groups 'everyone' and/or 'authenticated', which
        correspond to all users and all logged in users respectively.

        """
        raise NotImplementedError

    def user_permissions(self, package, username=None):
        """
        Get a mapping of all users to their permissions for a package

        If a username is specified, just return the list of permissions for
        that user

        Parameters
        ----------
        package : str
            The name of a python package
        username : str
            The name of a single user the check

        Returns
        -------
        permissions : dict
            Mapping of username to a list of permissions (which can contain
            'read' and/or 'write')
        permissions : list
            If username is not None, a list of permissions for that user

        """
        raise NotImplementedError

    def user_package_permissions(self, username):
        """
        Get a list of all packages that a user has permissions on

        Parameters
        ----------
        username : str

        Returns
        -------
        packages : list
            List of dicts. Each dict contains 'package' (str) and 'permissions'
            (list)

        """
        raise NotImplementedError

    def group_package_permissions(self, group):
        """
        Get a list of all packages that a group has permissions on

        Parameters
        ----------
        group : str

        Returns
        -------
        packages : list
            List of dicts. Each dict contains 'package' (str) and 'permissions'
            (list)

        """
        raise NotImplementedError

    def user_data(self, username=None):
        """
        Get a list of all users or data for a single user

        For Mutable backends, this MUST exclude all pending users

        Returns
        -------
        users : list
            Each user is a dict with a 'username' str, and 'admin' bool
        user : dict
            If a username is passed in, instead return one user with the fields
            above plus a 'groups' list.

        """
        raise NotImplementedError

    def dump(self):
        """
        Dump all of the access control data to a universal format

        Returns
        -------
        data : dict

        """
        from pypicloud import __version__
        data = {}
        data['allow_register'] = self.allow_register()
        data['version'] = __version__

        groups = self.groups()
        users = self.user_data()
        for user in users:
            user['password'] = self._get_password_hash(user['username'])

        data['groups'] = {}
        packages = {
            'users': defaultdict(dict),
            'groups': defaultdict(dict),
        }
        for group in groups:
            data['groups'][group] = self.group_members(group)
            perms = self.group_package_permissions(group)
            for perm in perms:
                package = perm['package']
                packages['groups'][package][group] = perm['permissions']

        for user in users:
            username = user['username']
            perms = self.user_package_permissions(username)
            for perm in perms:
                package = perm['package']
                packages['users'][package][username] = perm['permissions']

        # Convert the defaultdict to a dict for easy serialization
        packages['users'] = dict(packages['users'])
        packages['groups'] = dict(packages['groups'])
        data['users'] = users
        data['packages'] = packages
        return data

    def load(self, data):
        """
        Idempotently load universal access control data.

        By default, this does nothing on immutable backends. Backends may
        override this method to provide an implementation.

        This method works by default on mutable backends with no override
        necessary.

        """
        raise TypeError("Access backend '%s' is not mutable and has no "
                        "'load' implementation" % self.__class__.__name__)


class IMutableAccessBackend(IAccessBackend):

    """
    Base class for access backends that can change user/group permissions

    """
    mutable = True

    def need_admin(self):
        for user in self.user_data():
            if user['admin']:
                return False
        return True

    def allow_register(self):
        raise NotImplementedError

    def set_allow_register(self, allow):
        """
        Allow or disallow user registration

        Parameters
        ----------
        allow : bool

        """
        raise NotImplementedError

    def register(self, username, password):
        """
        Register a new user

        The new user should be marked as pending admin approval

        Parameters
        ----------
        username : str
        password : str
            This should be the plaintext password

        """
        if self.allow_register():
            self._register(username, pwd_context.encrypt(password))

    def _register(self, username, password):
        """
        Register a new user

        The new user should be marked as pending admin approval

        Parameters
        ----------
        username : str
        password : str
            This will be the hash of the password

        """
        raise NotImplementedError

    def pending_users(self):
        """
        Retrieve a list of all users pending admin approval

        Returns
        -------
        users : list
            List of usernames

        """
        raise NotImplementedError

    def approve_user(self, username):
        """
        Mark a user as approved by the admin

        Parameters
        ----------
        username : str

        """
        raise NotImplementedError

    def edit_user_password(self, username, password):
        """
        Change a user's password

        Parameters
        ----------
        username : str
        password : str

        """
        self._set_password_hash(username, pwd_context.encrypt(password))

    def _set_password_hash(self, username, password_hash):
        """
        Change a user's password

        Parameters
        ----------
        username : str
        password_hash : str
            The hashed password to store

        """
        raise NotImplementedError

    def delete_user(self, username):
        """
        Delete a user

        Parameters
        ----------
        username : str

        """
        raise NotImplementedError

    def set_user_admin(self, username, admin):
        """
        Grant or revoke admin permissions for a user

        Parameters
        ----------
        username : str
        admin : bool
            If True, grant permissions. If False, revoke.

        """
        raise NotImplementedError

    def edit_user_group(self, username, group, add):
        """
        Add or remove a user to/from a group

        Parameters
        ----------
        username : str
        group : str
        add : bool
            If True, add to group. If False, remove.

        """
        raise NotImplementedError

    def create_group(self, group):
        """
        Create a new group

        Parameters
        ----------
        group : str

        """
        raise NotImplementedError

    def delete_group(self, group):
        """
        Delete a group

        Parameters
        ----------
        group : str

        """
        raise NotImplementedError

    def edit_user_permission(self, package, username, perm, add):
        """
        Grant or revoke a permission for a user on a package

        Parameters
        ----------
        package : str
        username : str
        perm : {'read', 'write'}
        add : bool
            If True, grant permissions. If False, revoke.

        """
        raise NotImplementedError

    def edit_group_permission(self, package, group, perm, add):
        """
        Grant or revoke a permission for a group on a package

        Parameters
        ----------
        package : str
        group : str
        perm : {'read', 'write'}
        add : bool
            If True, grant permissions. If False, revoke.

        """
        raise NotImplementedError

    def dump(self):
        data = super(IMutableAccessBackend, self).dump()
        pending_users = []
        for username in self.pending_users():  # pylint: disable=E1101
            password = self._get_password_hash(username)
            pending_users.append({
                'username': username,
                'password': password,
            })
        data['pending_users'] = pending_users
        return data

    def load(self, data):
        # Have to temporarily set this as True for the load operation
        self.set_allow_register(True)
        pending_users = set(self.pending_users())

        def user_exists(username):
            """ Helper function that checks if a user already exists """
            return (username in pending_users or
                    self.user_data(username) is not None)

        for user in data['users']:
            if not user_exists(user['username']):
                self._register(user['username'], user['password'])
                self.approve_user(user['username'])
            self.set_user_admin(user['username'], user.get('admin', False))

        for group, members in data['groups'].iteritems():
            if len(self.group_members(group)) == 0:
                self.create_group(group)
            current_members = self.group_members(group)
            add_members = set(members) - set(current_members)
            for member in add_members:
                self.edit_user_group(member, group, True)

        for user in data.get('pending_users', []):
            if not user_exists(user['username']):
                self._register(user['username'], user['password'])

        for package, groups in data['packages']['groups'].iteritems():
            for group, permissions in groups.iteritems():
                for perm in permissions:
                    self.edit_group_permission(package, group, perm, True)

        for package, users in data['packages']['users'].iteritems():
            for user, permissions in users.iteritems():
                for perm in permissions:
                    self.edit_user_permission(package, user, perm, True)

        self.set_allow_register(data['allow_register'])



""" Caching database implementations """
from functools import partial
from pyramid.path import DottedNameResolver

from .base import ICache
from .redis_cache import RedisCache
from .sql import SQLCache


def includeme(config):
    """ Get and configure the cache db wrapper """
    settings = config.get_settings()
    resolver = DottedNameResolver(__name__)
    dotted_cache = settings.get('pypi.db', 'sql')
    if dotted_cache == 'sql':
        dotted_cache = 'pypicloud.cache.SQLCache'
    elif dotted_cache == 'redis':
        dotted_cache = 'pypicloud.cache.RedisCache'
    elif dotted_cache == 'dynamo':
        dotted_cache = 'pypicloud.cache.dynamo.DynamoCache'
    cache_impl = resolver.resolve(dotted_cache)
    kwargs = cache_impl.configure(settings)
    cache = cache_impl(**kwargs)
    cache.reload_if_needed()
    config.add_request_method(partial(cache_impl, **kwargs), name='db',
                              reify=True)
    return cache_impl



""" Store packages in S3 """
import calendar
import logging
import posixpath
import time
from contextlib import contextmanager
from hashlib import md5
from urllib import urlopen

import boto.s3
from boto.s3.key import Key
from pyramid.httpexceptions import HTTPFound
from pyramid.settings import asbool

from .base import IStorage
from pypicloud.models import Package
from pypicloud.util import parse_filename, getdefaults


LOG = logging.getLogger(__name__)


class S3Storage(IStorage):

    """ Storage backend that uses S3 """
    test = False

    def __init__(self, request=None, bucket=None, expire_after=None,
                 bucket_prefix=None, prepend_hash=None,
                 **kwargs):
        super(S3Storage, self).__init__(request, **kwargs)
        self.bucket = bucket
        self.expire_after = expire_after
        self.bucket_prefix = bucket_prefix
        self.prepend_hash = prepend_hash

    @classmethod
    def configure(cls, settings):
        kwargs = super(S3Storage, cls).configure(settings)
        kwargs['expire_after'] = int(getdefaults(
            settings, 'storage.expire_after', 'aws.expire_after', 60 * 60 *
            24))
        kwargs['bucket_prefix'] = getdefaults(
            settings, 'storage.prefix', 'aws.prefix', '')
        kwargs['prepend_hash'] = asbool(getdefaults(
            settings, 'storage.prepend_hash', 'aws.prepend_hash', True))
        access_key = getdefaults(settings, 'storage.access_key',
                                 'aws.access_key', None)
        secret_key = getdefaults(settings, 'storage.secret_key',
                                 'aws.secret_key', None)

        # We used to always use boto.connect_s3 because it can look up buckets
        # in any region. New regions require AWS4-HMAC-SHA256, which boto can
        # only do with a region connection. So if the region is specified (and
        # it must be for new regions like eu-central-1), use a region
        # connection.
        location = settings.get('storage.region')
        if location is None:
            s3conn = boto.connect_s3(
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key)
        else:
            s3conn = boto.s3.connect_to_region(location,
                                               aws_access_key_id=access_key,
                                               aws_secret_access_key=secret_key)
        aws_bucket = getdefaults(settings, 'storage.bucket', 'aws.bucket',
                                 None)
        if aws_bucket is None:
            raise ValueError("You must specify the 'storage.bucket'")
        try:
            bucket = s3conn.get_bucket(aws_bucket)
        except boto.exception.S3ResponseError as e:
            if e.error_code != 'NoSuchBucket':
                if e.status == 301:
                    LOG.warn("Bucket found in different region. Check that "
                             "the S3 bucket specified in 'storage.bucket' is "
                             "in 'storage.region'")
                raise
            location = getdefaults(settings, 'storage.region', 'aws.region',
                                   boto.s3.connection.Location.DEFAULT)
            LOG.info("Creating S3 bucket %s in region %s", aws_bucket,
                     location)
            bucket = s3conn.create_bucket(aws_bucket, location=location)
        kwargs['bucket'] = bucket
        return kwargs

    def get_path(self, package):
        """ Get the fully-qualified bucket path for a package """
        if 'path' not in package.data:
            filename = package.name + '/' + package.filename
            if self.prepend_hash:
                m = md5()
                m.update(package.filename)
                prefix = m.digest().encode('hex')[:4]
                filename = prefix + '/' + filename
            package.data['path'] = self.bucket_prefix + filename
        return package.data['path']

    def list(self, factory=Package):
        keys = self.bucket.list(self.bucket_prefix)
        for key in keys:
            # Moto doesn't send down metadata from bucket.list()
            if self.test:
                key = self.bucket.get_key(key.key)
            filename = posixpath.basename(key.key)
            name = key.get_metadata('name')
            version = key.get_metadata('version')

            # We used to not store metadata. This is for backwards
            # compatibility
            if name is None or version is None:
                try:
                    name, version = parse_filename(filename)
                except ValueError:
                    LOG.warning("S3 file %s has no package name", key.key)
                    continue

            last_modified = boto.utils.parse_ts(key.last_modified)

            pkg = factory(name, version, filename, last_modified, path=key.key)

            yield pkg

    def get_url(self, package):
        key = Key(self.bucket, self.get_path(package))
        return key.generate_url(self.expire_after)

    def download_response(self, package):
        return HTTPFound(location=self.get_url(package))

    def upload(self, package, data):
        key = Key(self.bucket)
        key.key = self.get_path(package)
        key.set_metadata('name', package.name)
        key.set_metadata('version', package.version)
        # S3 doesn't support uploading from a non-file stream, so we have to
        # read it into memory :(
        key.set_contents_from_string(data.read())

    def delete(self, package):
        path = self.get_path(package)
        key = Key(self.bucket)
        key.key = path
        key.delete()

    @contextmanager
    def open(self, package):
        url = self.get_url(package)
        handle = urlopen(url)
        try:
            yield handle
        finally:
            handle.close()
